/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   NAMES.C - Object name and type routines.
   */

/* the includes */
#include "deu.h"
#include <assert.h>

char *LookupKeyName (key_ptr *array, BCINT type, int numkeys);

sector_class *Sector_classes = NULL;
ld_class *Linedef_classes = NULL;
enum_type *Enum_classes = NULL;

ld_type *ld_array[256];

/*
   get the name of an object type
   */

char *GetObjectTypeName( BCINT objtype)
{
    switch (objtype) {
		
    case OBJ_THINGS:
		return "Thing";
    case OBJ_LINEDEFS:
		return "LineDef";
    case OBJ_SIDEDEFS:
		return "SideDef";
    case OBJ_VERTEXES:
		return "Vertex";
    case OBJ_SEGS:
		return "Segment";
    case OBJ_SSECTORS:
		return "SSector";
    case OBJ_NODES:
		return "Node";
    case OBJ_SECTORS:
		return "Sector";
    case OBJ_REJECT:
		return "Reject";
    case OBJ_BLOCKMAP:
		return "Blockmap";
    }
    return "< Bug! >";
}



/*
   what are we editing?
   */

char *GetEditModeName( BCINT objtype)
{
    switch (objtype) {
		
    case OBJ_THINGS:
		return "Things";
    case OBJ_LINEDEFS:
    case OBJ_SIDEDEFS:
		return "LineDefs & SideDefs";
    case OBJ_VERTEXES:
		return "Vertices";
    case OBJ_SEGS:
		return "Segments";
    case OBJ_SSECTORS:
		return "Seg-Sectors";
    case OBJ_NODES:
		return "Nodes";
    case OBJ_SECTORS:
		return "Sectors";
    }
    return "< Bug! >";
}


void index_ld_types (void)
{
	ld_class *c;
	ld_type *t;

	memset (ld_array, 0, sizeof(ld_array));

	for (c = Linedef_classes; c; c = c->next)
	{
		for (t = c->types; t; t = t->next)
		{
			if (t->type >= 0 && t->type <= 255)
			{
				ld_array[t->type] = t;
			}
		}
	}
}

int do_index_keys (key_type *keys, key_ptr **array)
{
	key_type *k;
	int i;

	// Count the keys
	for (i = 0, k = keys; k != NULL; ++i)
	{
		k = k->next;
	}
	*array = (key_ptr *)myalloc (i * sizeof(key_ptr));
	for (i = 0, k = keys; k != NULL; k = k->next)
	{
		if ((unsigned)k->type < 256)
		{
			(*array)[i].key = k;
			(*array)[i].type = k->type;
			++i;
		}
	}
	qsort(*array, i, sizeof(key_ptr), BCINT_lessp);
	return i;
}

void index_keys (void)
{
	enum_type *p;

	for (p = Enum_classes; p; p = p->next)
	{
		p->count = do_index_keys (p->keys, &p->array);
	}
}

enum_type *FindEnum (char *name)
{
	enum_type *p;

	if (name == NULL)
	{
		return NULL;
	}

	for (p = Enum_classes; p; p = p->next)
	{
		if (strcmp (p->name, name) == 0)
		{
			return p;
		}
	}
	return NULL;
}

char *GetEnumName (BCINT value, char *enumName)
{
	enum_type *e = FindEnum (enumName);
	if (e)
	{
		char *name = LookupKeyName (e->array, value, e->count);
		return name ? name : "";
	}
	return "";
}

char *GetUserFlagsName (BCINT value, char *flagsName)
{
	enum_type *e = FindEnum (flagsName);
	if (e)
	{
		static char temp[80];
		key_type *key;

		temp[0] = 0;
		for (key = e->keys; key != NULL; key = key->next)
		{
			if (key->name[0] != '-')
			{
				if (key->mask == 0)
				{
					if (value & key->type)
					{
						if (key->shortOn)
						{
							strcat (temp, key->shortOn);
						}
					}
					else
					{
						if (key->shortOff)
						{
							strcat (temp, key->shortOff);
						}
					}
				}
				else
				{
					if ((value & key->mask) == key->type)
					{
						if (key->shortOn)
						{
							strcat (temp, key->shortOn);
						}
					}
				}
			}
		}
		return temp;
	}
	return "";
}

/*
   get a short (2 + 13 char.) description of the type of a linedef
   */

char *GetLineDefTypeName( BCINT type)
{
	if ((unsigned short)type < 256 && ld_array[type])
	{
		return ld_array[type]->shortname;
	}
	return("Unknown Special");
}

/*
   get a long description of the type of a linedef
   */

char *GetLineDefTypeLongName( BCINT type)
{
	if ((unsigned short)type < 256 && ld_array[type])
	{
		return ld_array[type]->longname;
	}
	return("Unknown Special");
}

char *GetLineDefArgs( BCINT type)
{
	if ((unsigned short)type < 256 && ld_array[type])
	{
		return ld_array[type]->args;
	}
	return "-----";
}

char **GetArgText( BCINT type)
{
	static char *error[5] = { "ERROR", "ERROR", "ERROR", "ERROR", "ERROR" };
	if ((unsigned short)type < 256 && ld_array[type])
	{
		return ld_array[type]->argtext;
	}
	return error;
}

char **GetLineDefEnums (BCINT type)
{
	static char *error[5] = { 0 };
	if ((unsigned short)type < 256 && ld_array[type])
	{
		return ld_array[type]->enums;
	}
	return error;
}


/*
   get a short description of the flags of a linedef
   */

char *GetLineDefFlagsName( BCINT flags)
{
    static char temp[32];

	sprintf (temp, "%s%s%s%s%s%s%s%s%s%s%s%s",
		(flags & 0x8000) ? "Be" : "-", /* Blocks everything */
		(flags & 0x2000) ? "Am" : "-", /* Monsters can activate line */
		(flags & 0x0200) ? "Re" : "-", /* Repeatable */
		(flags & 0x0100) ? "Ma" : "-", /* Already on the map */
		(flags & 0x0080) ? "In" : "-", /* Invisible on the map */
		(flags & 0x0040) ? "So" : "-", /* Blocks sound */
		(flags & 0x0020) ? "Se" : "-", /* Secret (normal on the map) */
		(flags & 0x0010) ? "Lo" : "-", /* Lower texture offset changed */
		(flags & 0x0008) ? "Up" : "-", /* Upper texture offset changed */
		(flags & 0x0004) ? "2S" : "-", /* Two-sided */
		(flags & 0x0002) ? "Mo" : "-", /* Monsters can't cross this line */
		(flags & 0x0001) ? "Im" : "-"  /* Impassible */
	);
    return temp;
}


/*
   get a short description of the flags of a thing
   */

char *GetThingFlagsName( int flags, int extra)
{
    static char temp[26];

	sprintf (temp, "%c%c%c|%c%c%c|%c%c%c%s%s",
		(flags & 0x001) ? 'E' : '-',
		(flags & 0x002) ? 'M' : '-',
		(flags & 0x004) ? 'H' : '-',

		(flags & 0x020) ? 'F' : '-',
		(flags & 0x040) ? 'C' : '-',
		(flags & 0x080) ? 'M' : '-',

		(flags & 0x100) ? 'S' : '-',
		(flags & 0x200) ? 'C' : '-',
		(flags & 0x400) ? 'D' : '-',

		(flags & 0x008) ? ((extra & TX_SECTORACTION) ? " Monst" : " Deaf") : "",
		(flags & 0x010) ? ((extra & TX_SECTORACTION) ? " Missile" : " Dormant") : "");

    return temp;
}


/*
   get a long description of one linedef flag
   */

char *GetLineDefFlagsLongName( BCINT flags)
{
	if (flags & 0x8000)
		return "Blocks Everything";
	if (flags & 0x2000)
		return "Activateable by both Monsters and Players";
    if (flags & 0x0200)
		return "Repeatable";
    if (flags & 0x0100)
		return "Already On The Map At Startup";
    if (flags & 0x80)
		return "Invisible On The Map";
    if (flags & 0x40)
		return "Blocks Sound";
    if (flags & 0x20)
		return "Secret (Normal On Map)";
    if (flags & 0x10)
		return "Lower Texture Is \"Unpegged\"";
    if (flags & 0x08)
		return "Upper Texture Is \"Unpegged\"";
    if (flags & 0x04)
		return "Two-Sided (May Be Transparent)";
    if (flags & 0x02)
		return "Monsters Cannot Cross This Line";
    if (flags & 0x01)
		return "Impassible";
    return "Unknown";
}

/*
   get a long description of one thing flag
   */

char *GetThingFlagLongName( int flags, int extra)
{
    if (flags & 0x0400)
		return "Deathmatch";
    if (flags & 0x0200)
		return "Cooperative";
    if (flags & 0x0100)
		return "Single-player";
    if (flags & 0x80)
		return "Mage";
    if (flags & 0x40)
		return "Cleric";
    if (flags & 0x20)
		return "Fighter";
    if (flags & 0x10)
		return (extra & TX_SECTORACTION) ? "Missiles Can Trigger" : "Dormant";
    if (flags & 0x08)
		return (extra & TX_SECTORACTION) ? "Monsters Can Trigger" : "Deaf";
    if (flags & 0x04)
		return "Hard";
    if (flags & 0x02)
		return "Medium";
    if (flags & 0x01)
		return "Easy";
    return "Unknown";
}

char *GetSectorFlags (BCINT special)
{
	static char foo[32];

	foo[0] = 0;
	switch (special & 0x0300)
	{
	case 0x0100: strcpy (foo, "D5 ");  break;
	case 0x0200: strcpy (foo, "D10 "); break;
	case 0x0300: strcpy (foo, "D20 "); break;
	}
	if (special & 0x0400)
		strcat (foo, "Secret ");
	if (special & 0x0800)
		strcat (foo, "Frict ");
	if (special & 0x1000)
		strcat (foo, "Wind");

	return foo;
}

/*
   get a short (19 char.) description of the type of a sector
   */

char *GetSectorTypeName (BCINT type)
{
	sector_class *c;
	sector_type *t;
	
	type &= 255;

	for(c = Sector_classes; c; c = c->next)
		for(t = c->types; t; t = t->next)
			if(type == t->type)
				return t->shortname;
	
    return "Unknown Sector!";
}



/*
   get a long description of the type of a sector
   */

char *GetSectorTypeLongName( BCINT type)
{
	sector_class *c;
	sector_type *t;
	
	type &= 255;

	for(c = Sector_classes; c; c = c->next)
		for(t = c->types; t; t = t->next)
			if(type == t->type)
				return t->longname;
	
    return "Unknown Sector!";
}

BCINT KeyMenu (key_type *keys, char *label)
{
	key_type *type;
	int val, i;
	char *menuarr[MENUARRAYMAX];
	char menuitems[MENUARRAYMAX][80];
	BCINT dummy[MENUARRAYMAX];
	int width = 0;

	for (type = keys; type != NULL; type = type->next)
	{
		int len = strlen (type->name);
		if (len > width)
		{
			width = len;
		}
	}
	if (width == 0)
	{
		return InputIntegerValue (-1, -1, 0, 255, 0);
	}
	if (width <= 17)
	{
		width = 21;
	}
	else
	{
		width += 4;
	}
	for (i = 0, type = keys; type != NULL && i < MENUARRAYMAX-2; ++i, type = type->next)
	{
		menuarr[i] = menuitems[i];
		if (menuitems[i][0] != '-')
		{
			sprintf (menuitems[i], "%-*s[%3d]", width, type->name, type->type);
		}
		else
		{
			menuitems[i][0] = '-';
			menuitems[i][1] = '\0';
		}
	}
	menuarr[i++] = "-";
	menuarr[i++] = "Enter a Decimal Value";
	menuarr[i++] = "-";
	menuarr[i++] = "Exit Menu";
	val = DisplayMenuArray (-1, -1, label, i, NULL, menuarr, dummy);

	if (val <= 0 || val >= i)
	{
		return -1;
	}
	if (val == i-2)
	{
		return InputIntegerValue (-1, -1, 0, 255, 0);
	}
	for (type = keys; type != NULL && --val; type = type->next)
	{
	}
	if ((unsigned short)type->type > 255)
	{
		return -1;
	}
	return type->type;
}

int TwiddleFlagsMenu (key_type *keys, char *label, int start, int *mask)
{
	int twiddle;
	key_type *type;
	int val, i, pos;
	char *menuarr[MENUARRAYMAX];
	char menuitems[MENUARRAYMAX][80];
	BCINT dummy[MENUARRAYMAX];
	int width = 0;

	for (type = keys; type != NULL; type = type->next)
	{
		int len = strlen (type->name);
		if (len > width)
		{
			width = len;
		}
	}
	if (width == 0)
	{
		return InputIntegerValue (-1, -1, 0, 255, 0);
	}
	if (width <= 17)
	{
		width = 17;
	}
	for (i = 0, type = keys; type != NULL && i < MENUARRAYMAX-2; ++i, type = type->next)
	{
		menuarr[i] = menuitems[i];
		if (type->name[0] != '-')
		{
			if (type->mask == 0)
			{
				sprintf (menuitems[i], "%c %-*s", (start & type->type) ? '*' : ' ', width, type->name);
			}
			else
			{
				sprintf (menuitems[i], "%c %-*s", ((start & type->mask) == type->type) ? '*' : ' ', width, type->name);
			}
		}
		else
		{
			menuarr[i][0] = '-';
			menuarr[i][1] = '\0';
		}
	}
	menuarr[i++] = "-";
	menuarr[i++] = "Enter a Decimal Value";
	menuarr[i++] = "-";
	menuarr[i++] = "Exit Menu";
	MenuDefault = 0;

	*mask = ~0;
	twiddle = start;

	for (;;)
	{
		val = DisplayMenuArray (-1, -1, label, i, NULL, menuarr, dummy);
		pos = val - 1;

		if (val <= 0 || val >= i)
		{
			twiddle ^= (start & *mask);
			return twiddle;
		}
		if (val == i-2)
		{
			int i;
			twiddle = InputIntegerValue (-1, -1, 0, 255, twiddle);
			*mask = 0;

refreshlist:
			for (i = 0, type = keys; type != NULL && i < MENUARRAYMAX-2; ++i, type = type->next)
			{
				if (type->name[0] != '-')
				{
					if (type->mask == 0)
					{
						menuitems[i][0] = (twiddle & type->type) ? '*' : ' ';
					}
					else
					{
						menuitems[i][0] = ((twiddle & type->mask) == type->type) ? '*' : ' ';
					}
				}
			}
			continue;
		}
		for (type = keys; type != NULL && --val; type = type->next)
		{
		}
		if (type == NULL || (unsigned short)type->type > 255)
		{
			twiddle ^= (start & *mask);
			return twiddle;
		}
		MenuDefault = pos;
		if (type->mask == 0)
		{
			twiddle ^= type->type;
			menuitems[pos][0] = (menuitems[pos][0] == '*') ? ' ' : '*';
		}
		else
		{
			*mask &= ~type->mask;
			twiddle = (twiddle & ~type->mask) | type->type;
			goto refreshlist;
		}
	}
	twiddle ^= (start & *mask);
	return twiddle;
}

BCINT SelectEnumValue(char *enumName)
{
	enum_type *e = FindEnum (enumName);
	char title[100];

	if (e == NULL)
	{
		return InputIntegerValue (-1, -1, 0, 255, 0);
	}
	sprintf (title, "Choose %s Value:", e->name);
	return KeyMenu (e->keys, title);
}

BCINT SelectHexenAngle()
{
	int val;
	static const BCINT angles[8] = { 64, 32, 0, 224, 192, 160, 128, 96 };

	val = DisplayMenu( -1, -1, "Choose Angle:",
		"North             [ 64]",
		"NorthEast         [ 32]",
		"East              [  0]",
		"SouthEast         [224]",
		"South             [192]",
		"SouthWest         [160]",
		"West              [128]",
		"NorthWest         [ 96]",
		"-",
		"Enter a Decimal Value",
		NULL);

	if (val >= 1 && val <= 8)
	{
		return angles[val-1];
	}
	else if (val == 10)
	{
		return InputIntegerValue (-1, -1, 0, 255, 0);
	}
	else
	{
		return -1;
	}
	return val;
}

BCINT SelectNegative()
{
	int val;
	val = DisplayMenu( -1, -1, "Choose Yes/No Value:",
		"Yes               [  1]",
		"No                [  0]",
		"-",
		"Enter a Decimal Value",
		NULL);
	switch (val) {
	    case 1:
		    val = 1;
		    break;
	    case 2:
		    val = 0;
		    break;
		case 3:
			return -1;
	    case 4:
		val = InputIntegerValue( -1, -1, 0, 255, 0);
	}
	return val;
}

char *LookupKeyName (key_ptr *array, BCINT type, int numkeys)
{
	int low = 0, high = numkeys-1;

	do
	{
		int mid = (low + high) / 2;
		int t = array[mid].type;
		if (t == type)
		{
			return array[mid].key->name;
		}
		else if (t > type)
		{
			high = mid - 1;
		}
		else
		{
			low = mid + 1;
		}
	} while (low <= high);
	return NULL;
}

char *GetNegativeName( BCINT val)
{
	return val == 0 ? "No" : "Yes";
}

char *GetHexenAngleName( BCINT val)
{
	static char *angleNames[8] =
	{
		"East",
		"NorthEast",
		"North",
		"NorthWest",
		"West",
		"SouthWest",
		"South",
		"SouthEast"
	};
	static char arbitraryName[12];

	if (val & 31)
	{
		sprintf (arbitraryName, "%d degrees", ((val&255)+64)*180/128);
		return arbitraryName;
	}
	return angleNames[val>>5];
}

/* end of file */

